@@ -1,21 +1,8 @@
 from datetime import datetime
 from typing import Any, List, Tuple
 
-from uplink import (
-    Body,
-    Query,
-    delete,
-    get,
-    json,
-    params,
-    post,
-    put,
-    response_handler,
-    returns,
-)
-
-from bpkio_api.consumer import BpkioSdkConsumer
 from bpkio_api.caching import cache_api_results
+from bpkio_api.consumer import BpkioSdkConsumer
 from bpkio_api.exceptions import BroadpeakIoHelperError, ResourceExistsError
 from bpkio_api.helpers.list import collect_from_ids, get_all_with_pagination
 from bpkio_api.helpers.objects import find_duplicates_of
@@ -24,6 +11,8 @@ from bpkio_api.models import ContentReplacementSlot
 from bpkio_api.models import Services as Svc
 from bpkio_api.models import VirtualChannelSlot, VirtualChannelSlotIn
 from bpkio_api.response_handler import postprocess_response, return_count
+from uplink import (Body, Query, delete, get, json, params, post, put,
+                    response_handler, returns)
 
 from .enums import UpsertOperationType
 
@@ -42,6 +31,9 @@ class ServicesApi(BpkioSdkConsumer):
         self.ad_insertion = AdInsertionServiceApi(
             parent_api=self, base_url=base_url, **kwargs
         )
+        self.adaptive_streaming_cdn = AdaptiveStreamingCdnServiceApi(
+            parent_api=self, base_url=base_url, **kwargs
+        )
 
     def _mappings(self, model):
         match model:
@@ -57,6 +49,8 @@ class ServicesApi(BpkioSdkConsumer):
                 return self.content_replacement
             case Svc.ContentReplacementService():
                 return self.content_replacement
+            case Svc.AdaptiveStreamingCdnService():
+                return self.adaptive_streaming_cdn
             case _:
                 raise Exception(
                     f"Model {model.__class__.__name__} "
@@ -118,6 +112,7 @@ class ServicesApi(BpkioSdkConsumer):
         Svc.AdInsertionService
         | Svc.ContentReplacementService
         | Svc.VirtualChannelService
+        | Svc.AdaptiveStreamingCdnService
         | None
     ):
         """Gets a service by its ID
@@ -148,9 +143,10 @@ class ServicesApi(BpkioSdkConsumer):
                 return self.virtual_channel.retrieve(service_id)
             if service.type == Svc.ServiceType.CONTENT_REPLACEMENT:
                 return self.content_replacement.retrieve(service_id)
+            if service.type == Svc.ServiceType.ADAPTIVE_STREAMING_CDN:
+                return self.adaptive_streaming_cdn.retrieve(service_id)
         except IndexError as e:
             raise BroadpeakIoHelperError(
-                status_code=404,
                 message=f"There is no service with ID {service_id}",
                 original_message=e.args[0],
             )
@@ -171,7 +167,6 @@ class ServicesApi(BpkioSdkConsumer):
         service = self.retrieve(service_id)
         if not service:
             raise BroadpeakIoHelperError(
-                status_code=404,
                 message=f"There is no service with ID {service_id}",
             )
         endpoint = self._mappings(service)
@@ -255,6 +250,32 @@ class ServicesApi(BpkioSdkConsumer):
                         updated_resource = self._update(existing_resource.id, service)
                         return (updated_resource, UpsertOperationType.UPDATED)
 
+    def pause(
+        self, service_id: int
+    ) -> (
+        Svc.VirtualChannelService
+        | Svc.ContentReplacementService
+        | Svc.AdInsertionService
+    ):
+        """Disable (pause) a service"""
+        service = self.retrieve(service_id)
+        service.state = "paused"
+        endpoint = self._mappings(service)
+        return endpoint.update(service_id, service)
+
+    def unpause(
+        self, service_id: int
+    ) -> (
+        Svc.VirtualChannelService
+        | Svc.ContentReplacementService
+        | Svc.AdInsertionService
+    ):
+        """Enable (unpause) a service"""
+        service = self.retrieve(service_id)
+        service.state = "enabled"
+        endpoint = self._mappings(service)
+        return endpoint.update(service_id, service)
+
 
 # === CONTENT-REPLACEMENT SERVICES ===
 
@@ -298,6 +319,14 @@ class ContentReplacementServiceApi(BpkioSdkConsumer):
     def delete(self, service_id: int):
         """Delete a Content Replacement service, by ID"""
 
+    def pause(self, service_id: int) -> Svc.ContentReplacementService:
+        """Disable (pause) a Content Replacement service, by ID"""
+        return self.parent_api.pause(service_id)
+
+    def unpause(self, service_id: int) -> Svc.ContentReplacementService:
+        """Enable (unpause) a Content Replacement service, by ID"""
+        return self.parent_api.unpause(service_id)
+
     @cache_api_results("list_cr")
     def list(self) -> List[Svc.ContentReplacementService]:
         """List all Content Replacement services"""
@@ -322,6 +351,25 @@ class ContentReplacementServiceApi(BpkioSdkConsumer):
         services = self.list()
         return search_array_with_filters(services, filters=filters)
 
+    def clear(self, service_id):
+        """Delete all Content Replacement slots, for a given service
+
+        Args:
+            service_id (int): ID of the Content Replacement service
+
+        Returns:
+            Tuple: Number of slots successfully and unsuccessfully deleted
+        """
+        deleted = 0
+        failed = 0
+        for slot in self.slots.list(service_id):
+            try:
+                self.slots.delete(service_id, slot.id)
+                deleted += 1
+            except Exception:
+                failed += 1
+        return (deleted, failed)
+
 
 @response_handler(postprocess_response)
 class ContentReplacementServiceSlotsApi(BpkioSdkConsumer):
@@ -347,25 +395,6 @@ class ContentReplacementServiceSlotsApi(BpkioSdkConsumer):
     def delete(self, service_id, slot_id):
         """Delete a Content Replacement slot, by ID"""
 
-    def clear(self, service_id):
-        """Delete all Content Replacement slots for a service
-
-        Args:
-            service_id (_type_): ID of the Content Replacement service
-
-        Returns:
-            Tuple: Number of slots successfully and unsuccessfully deleted
-        """
-        deleted = 0
-        failed = 0
-        for slot in self.list(service_id):
-            try:
-                self.delete(service_id, slot.id)
-                deleted += 1
-            except Exception:
-                failed += 1
-        return (deleted, failed)
-
     @response_handler(return_count)
     @params({"offset": 0, "limit": 1})
     @get("services/content-replacement/{service_id}/slots")
@@ -383,16 +412,20 @@ class ContentReplacementServiceSlotsApi(BpkioSdkConsumer):
         service_id: int,
         from_time: datetime | None = None,
         to_time: datetime | None = None,
-        categories: List[int] = [],
+        categories: List[int] | None = [],
     ) -> List[ContentReplacementSlot]:
         """Get the full list of Content Replacement slots"""
-        return get_all_with_pagination(
+        slots = get_all_with_pagination(
             self.get_page,
             service_id=service_id,
             from_time=from_time,
             to_time=to_time,
             categories=categories,
         )
+        if categories is None:
+            return [s for s in slots if s.category is None]
+        else:
+            return slots
 
 
 # === VIRTUAL-CHANNEL SERVICES ===
@@ -437,6 +470,14 @@ class VirtualChannelServiceApi(BpkioSdkConsumer):
     def delete(self, service_id: int):
         """Delete a Virtual Channel service, by ID"""
 
+    def pause(self, service_id: int):
+        """Pause a Virtual Channel service, by ID"""
+        return self.parent_api.pause(service_id)
+
+    def unpause(self, service_id: int):
+        """Unpause a Virtual Channel service, by ID"""
+        return self.parent_api.unpause(service_id)
+
     @cache_api_results("list_vcs")
     def list(self) -> List[Svc.VirtualChannelService]:
         """List all Virtual Channel services"""
@@ -459,6 +500,25 @@ class VirtualChannelServiceApi(BpkioSdkConsumer):
         services = self.list()
         return search_array_with_filters(services, filters=filters)
 
+    def clear(self, service_id):
+        """Delete all Virtual Channel slots, for a given service
+
+        Args:
+            service_id (int): ID of the Content Replacement service
+
+        Returns:
+            Tuple: Number of slots successfully and unsuccessfully deleted
+        """
+        deleted = 0
+        failed = 0
+        for slot in self.slots.list(service_id):
+            try:
+                self.slots.delete(service_id, slot.id)
+                deleted += 1
+            except Exception:
+                failed += 1
+        return (deleted, failed)
+
 
 @response_handler(postprocess_response)
 class VirtualChannelServiceSlotsApi(BpkioSdkConsumer):
@@ -492,25 +552,6 @@ class VirtualChannelServiceSlotsApi(BpkioSdkConsumer):
     def delete(self, service_id, slot_id):
         """Delete a Virtual Channel slot, by ID"""
 
-    def clear(self, service_id):
-        """Delete all Virtual Channel slots, for a given service
-
-        Args:
-            service_id (int): ID of the Content Replacement service
-
-        Returns:
-            Tuple: Number of slots successfully and unsuccessfully deleted
-        """
-        deleted = 0
-        failed = 0
-        for slot in self.list(service_id):
-            try:
-                self.delete(service_id, slot.id)
-                deleted += 1
-            except Exception:
-                failed += 1
-        return (deleted, failed)
-
     @response_handler(return_count)
     @params({"offset": 0, "limit": 1})
     @get("services/virtual-channel/{service_id}/slots")
@@ -528,16 +569,20 @@ class VirtualChannelServiceSlotsApi(BpkioSdkConsumer):
         service_id: int,
         from_time: datetime = None,
         to_time: datetime = None,
-        categories: List[int] = [],
+        categories: List[int] | None = [],
     ) -> List[VirtualChannelSlot]:
         """Get the full list of Virtual Channel slots"""
-        return get_all_with_pagination(
+        slots = get_all_with_pagination(
             self.get_page,
             service_id=service_id,
             from_time=from_time,
             to_time=to_time,
             categories=categories,
         )
+        if categories is None:
+            return [s for s in slots if s.category is None]
+        else:
+            return slots
 
 
 # === AD-INSERTION SERVICES ===
@@ -580,6 +625,14 @@ class AdInsertionServiceApi(BpkioSdkConsumer):
     def delete(self, service_id: int):
         """Delete an Ad Insertion service, by ID"""
 
+    def pause(self, service_id: int):
+        """Pause an Ad Insertion service, by ID"""
+        return self.parent_api.pause(service_id)
+
+    def unpause(self, service_id: int):
+        """Unpause an Ad Insertion service, by ID"""
+        return self.parent_api.unpause(service_id)
+
     @cache_api_results("list_dai")
     def list(self) -> List[Svc.AdInsertionService]:
         """List all Ad Insertion services"""
@@ -601,3 +654,76 @@ class AdInsertionServiceApi(BpkioSdkConsumer):
 
         services = self.list()
         return search_array_with_filters(services, filters=filters)
+
+
+# === ADADTIVE-STREAMING-CDN SERVICES ===
+
+
+@response_handler(postprocess_response)
+class AdaptiveStreamingCdnServiceApi(BpkioSdkConsumer):
+    def __init__(self, parent_api: ServicesApi, base_url="", **kwargs):
+        super().__init__(base_url, **kwargs)
+        self.parent_api = parent_api
+
+    @returns.json(Svc.AdaptiveStreamingCdnService)
+    @get("services/adaptive-streaming-cdn/{service_id}")
+    def retrieve(self, service_id):
+        """Get a single Adaptive Streaming CDN service, by ID"""
+
+    @json
+    @returns.json(Svc.AdaptiveStreamingCdnService)
+    @post("services/adaptive-streaming-cdn")
+    def create(
+        self, service: Body(type=Svc.AdaptiveStreamingCdnServiceIn)  # type: ignore
+    ) -> Svc.AdaptiveStreamingCdnService:  # type: ignore
+        """Create a new Adaptive Streaming CDN service"""
+
+    @json
+    @returns.json(Svc.AdaptiveStreamingCdnService)
+    @put("services/adaptive-streaming-cdn/{service_id}")
+    def update(self, service_id: int, service: Body(type=Svc.AdaptiveStreamingCdnServiceIn)) -> Svc.AdaptiveStreamingCdnService:  # type: ignore
+        """Update an Adaptive Streaming CDN service"""
+
+    def upsert(
+        self, service: Svc.AdaptiveStreamingCdnServiceIn, if_exists: str = "retrieve"
+    ) -> Tuple[Svc.AdaptiveStreamingCdnService, UpsertOperationType]:
+        """Conditionally create, retrieve or update an Adaptive Streaming CDN service"""
+        return self.parent_api.upsert(
+            service, unique_fields=["url"], if_exists=if_exists
+        )
+
+    @delete("services/adaptive-streaming-cdn/{service_id}")
+    def delete(self, service_id: int):
+        """Delete an Adaptive Streaming CDN service, by ID"""
+
+    def pause(self, service_id: int):
+        """Pause an Adaptive Streaming CDN service, by ID"""
+        return self.parent_api.pause(service_id)
+
+    def unpause(self, service_id: int):
+        """Unpause an Adaptive Streaming CDN service, by ID"""
+        return self.parent_api.unpause(service_id)
+
+    @cache_api_results("list_ascdn")
+    def list(self) -> List[Svc.AdaptiveStreamingCdnService]:
+        """List all Adaptive Streaming CDN services"""
+        sparse_vcs = self.parent_api.search(
+            Svc.ServiceType.ADAPTIVE_STREAMING_CDN,
+            field="type",
+            method=SearchMethod.STRICT,
+        )
+        return collect_from_ids(ids=[vc.id for vc in sparse_vcs], get_fn=self.retrieve)
+
+    def search(
+        self,
+        value: Any | None = None,
+        field: str | None = None,
+        method: SearchMethod = SearchMethod.STRING_SUB,
+        filters: List[Tuple[Any, str | None, SearchMethod | None]] | None = None,
+    ) -> List[Svc.AdaptiveStreamingCdnService]:
+        """Searches the list of Adaptive Streaming CDN services for those matching a particular filter query"""
+        if not filters:
+            filters = [(value, field, method)]
+
+        services = self.list()
+        return search_array_with_filters(services, filters=filters)
